#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "carpet.hh"
#define RHOGF 6.1755e17

extern "C" {
  int CoreCollapseControl_PMRStartup(void);
  void CoreCollapseControl_PMRInit(CCTK_ARGUMENTS);
  void CoreCollapseControl_PMRGetRhoMax(CCTK_ARGUMENTS);
  void CoreCollapseControl_PMRCheckRegrid(CCTK_ARGUMENTS);
}
  

int CoreCollapseControl_PMRStartup(void) 
{
  const char *banner = "CoreCollapseControl: Progressive Mesh Refinement";
  
  CCTK_RegisterBanner(banner);
  
  return 0;
  
}

void CoreCollapseControl_PMRInit(CCTK_ARGUMENTS) 
{
  DECLARE_CCTK_ARGUMENTS;
  *rho_max_index = 0;
  *global_rho_max = 0.0e0;
  *global_entropy_max = 0.0e0;
  *rho_max = 0.0e0;
  *in_prebounce = 0;

  *bounce = 0;
  *bouncetime = 0.0e0;

  *global_alp_min = 10.0e0;
  *alp_min = 10.0e0;
  *in_preBH = 0;
  *force_check = 0;

  return;
}

void CoreCollapseControl_PMRGetRhoMax(CCTK_ARGUMENTS)
{

  DECLARE_CCTK_PARAMETERS;
  DECLARE_CCTK_ARGUMENTS;

  int reduction_handle, varindex, error, vartype;
  CCTK_REAL current_rho_max;
  CCTK_REAL current_alp_min;

  if(  (( (cctk_iteration) % rho_max_every) != 0) && !*force_check ) 
    return;

  varindex = CCTK_VarIndex("hydrobase::rho");
  if(varindex < 0) {
    CCTK_WARN(0,"Variable hydrobase::rho does not exist! Can't regrid");
  }    

  vartype = CCTK_VarTypeI(varindex);
  reduction_handle = CCTK_ReductionHandle("maximum");
  error = CCTK_Reduce(cctkGH,-1,reduction_handle,1,
		      vartype,&current_rho_max,1,varindex);

  if(error != 0) {
      CCTK_WARN(0,"Could not perform reduction on rho_max, no regridding.");
  }

  if(current_rho_max >= *rho_max) {
      *rho_max = current_rho_max;
  }

  if(*rho_max >= *global_rho_max) {
    *global_rho_max = *rho_max;
  }

  varindex = CCTK_VarIndex("admbase::alp");
  if(varindex < 0) {
    CCTK_WARN(0,"Variable admbase::alp does not exist! Can't steer parameters based on its values.");
  }    
  vartype = CCTK_VarTypeI(varindex);
  reduction_handle = CCTK_ReductionHandle("minimum");
  error = CCTK_Reduce(cctkGH,-1,reduction_handle,1,
		      vartype,&current_alp_min,1,varindex);

  if(error != 0) {
      CCTK_WARN(0,"Could not perform reduction on alp_min, no changes to AH/output frequency.");
  }

  *alp_min = current_alp_min;
  
  if(*alp_min < *global_alp_min) {
    *global_alp_min = *alp_min;
  }


  return;
} 

void CoreCollapseControl_PMRCheckRegrid(CCTK_ARGUMENTS)
{

  DECLARE_CCTK_PARAMETERS;
  DECLARE_CCTK_ARGUMENTS;

  int reduction_handle, varindex, error, vartype;
  CCTK_REAL current_rho_max;

  if( ( (cctk_iteration) % check_every) != 0 ) 
    return;

  current_rho_max = *rho_max;

  // don't do anything if we don't have a density set at this rho_max_index
  if(rho_max_list[*rho_max_index] < 1.0e-10) return;

  if (current_rho_max >= rho_max_list[*rho_max_index]/RHOGF) {
    // loop over all provided densities and check if they are above threshold
    // (this way we can switch on multiple levels at once!)
    while (current_rho_max >= rho_max_list[*rho_max_index]/RHOGF) {
      if(num_levels[0] < Carpet::maxreflevels) {
        (*rho_max_index)++;
        num_levels[0]++;

        // regrid by changing CarpetRegrid2 parameters
        CCTK_VInfo(CCTK_THORNSTRING,"CarpetRegrid2 has been told to regrid!!!!");
	if(rho_max_list[*rho_max_index] > 0.0e0) {
	  CCTK_VInfo(CCTK_THORNSTRING,"Next regrid at rho_max = %5.5E",
		     rho_max_list[*rho_max_index]);
	} else {
	  CCTK_Info(CCTK_THORNSTRING,"No further regridding planned");
	}
	  //       fprintf(stderr,"max rls: %d, cur rls: %d\n", Carpet::maxreflevels, num_levels[0]);
      }
    }
  } else {
    CCTK_VInfo(CCTK_THORNSTRING,
	       "Sorry, no regridding! rho max: %5.5E g/cm^3",current_rho_max*RHOGF);
    CCTK_VInfo(CCTK_THORNSTRING,"Next regrid at rho_max = %5.5E g/cm^3",
	       rho_max_list[*rho_max_index]);

    //    fprintf(stderr,"max rls: %d, cur rls: %d\n", Carpet::maxreflevels, num_levels[0]);
  }

  // in case we are beyond a given time, we may want to swictch off finer levels
  if (switch_off_refinement && cctk_time > switch_off_refinement_after_time)
  {
     num_levels[0] = 1;

     // regrid by changing CarpetRegrid2 parameters
     CCTK_VInfo(CCTK_THORNSTRING,"CarpetRegrid2 has been told to SWITCH OFF REFINEMENT!!!!");
  }

  return;
} 


